import os
import logging
import time
from concurrent import futures
from multiprocessing import Process
from queue import Queue
from threading import Thread

import grpc
import pickle
import datetime

from .protos import DataProducerServicer, \
	add_DataProducerServicer_to_server
from .protos import BatchData, BatchDataResponse
from .protos.data_consumer_pb2 import ProducerStopResponse, Hints

_MAX_SIZE_OF_MESSAGE = 4000000
default_hints = {"topic": "default", "units": 1}
logger = logging.getLogger('CPU_PRODUCER')


class _DataProducerServicer(DataProducerServicer):
	# extra bytes for reserving
	# _EXTRA_SPACE = 1000

	def __init__(self, data_queue: Queue, server):
		logger.info("Init _DataProducerServicer")
		self.data_queue = data_queue
		self.server = server
		self.previous_picked_data = None
		self.previous_hints = None

	def PullBatch(self, request_iterator, context):
		logger.info("Start pull batch grpc channel")
		for _ in request_iterator:
			logger.info("Get request")
			yield self._create_response()

	def Stop(self, request, context):
		process = Thread(target=_server_stop, args=(self.server,))
		process.start()
		return ProducerStopResponse()

	def _create_response(self):
		if self.previous_picked_data is None:
			logger.info("Previous data is none")
			batch = self.data_queue.get()
			hints = getattr(batch, "hints", default_hints)
			logger.info(f"Get batch with hints {hints}")
			self.previous_picked_data = pickle.dumps(
				batch
			)
			self.previous_hints = hints

		pickled_data = self.previous_picked_data
		is_final = True

		# with empty data it is equal to 17
		response = BatchDataResponse(
			is_final=is_final,
			data=BatchData(
				pickled_data=pickled_data
			),
			hints=Hints(
				units=self.previous_hints.get("units", 1),
				topic=self.previous_hints.get("topic", "default")
			)
		)
		logger.info(f"Size of message is equal to {response.ByteSize()}, max size is equals to {_MAX_SIZE_OF_MESSAGE}")
		if response.ByteSize() < _MAX_SIZE_OF_MESSAGE:
			self.previous_picked_data = None
			self.previous_hints = None
			return response
		excess = response.ByteSize() - _MAX_SIZE_OF_MESSAGE
		n_bytes = len(self.previous_picked_data)
		offset = n_bytes - excess
		logger.info(f"excess = {excess}, n_bytes = {n_bytes}, offset = {offset}")
		pickled_data = self.previous_picked_data[:offset]
		self.previous_picked_data = self.previous_picked_data[offset:]
		is_final = False
		return BatchDataResponse(
			is_final=is_final,
			data=BatchData(
				pickled_data=pickled_data
			),
			hints=Hints(
				units=self.previous_hints.get("units", 1),
				topic=self.previous_hints.get("topic", "default")
			)
		)


def _server_stop(server):
	time.sleep(5)
	server.stop()


def _serve(data_queue: Queue):
	port = int(os.environ['SERVER_PORT'])
	server = grpc.server(futures.ThreadPoolExecutor(max_workers=1))
	add_DataProducerServicer_to_server(
		servicer=_DataProducerServicer(data_queue, server),
		server=server
	)
	logger.info(f"Start server on port: {port}")
	server.add_insecure_port(f'[::]:{port}')
	server.start()


class _TaasDataIterator:

	def __init__(self, data_loader, data_queue: Queue):
		self._inner_iter = iter(data_loader)
		self._data_queue = data_queue

	def __iter__(self):
		return self

	def __next__(self):
		logger.info(f"next method call")
		time_before_getting = datetime.datetime.now()
		batch = next(self._inner_iter)
		print(f"Delta time = {datetime.datetime.now() - time_before_getting}")
		self._data_queue.put(batch)


def data_sender(data_loader):
	batch_queue = Queue(int(os.environ['MAX_QUEUE_LENGTH']))
	Process(target=_serve(batch_queue)).start()
	logger.info(f"After server starts")
	return _TaasDataIterator(
		data_loader=data_loader,
		data_queue=batch_queue
	)
